;;;
;;; solve.scm --- a semi-intelligent simple equation solver
;;;

;; Solves the problem e = a1*...*an*i where e is the identity, ai is
;; some given element, and i is the initial element.

(define not-seen (list 'not-seen))

(define (solve-to-identity g ops init)
  (solve-to-element g (group-identity g) ops init))

(define (solve-to-element g i ops init)
  (do-solve-to-element (has-tag? g 'inverse) g i ops init))

(define (do-solve-to-element bidirectional? g i ops init)
  (let* ((inv-op (if bidirectional? (group-inverse-operation g) (lambda (x) x))) ; lambda never used
         (inv-ops (map (lambda (e) (inv-op e)) ops))
         (seen (make-equal-hash-table)) ; prod,dir => op,parent
         (queue (make-queue)))
    (call-with-current-continuation
     (lambda (found-it)
       (hash-table/put! seen (cons init #t) (cons '() '()))
       (hash-table/put! seen (cons i #f) (cons '() '()))
       (if bidirectional?
           (enqueue!/unsafe queue (cons i #f)))
       (let lp ((prod (cons init #t)))
         (hash-table/lookup seen (cons (car prod) (not (cdr prod)))
                            (lambda (found-seen)
                              ((if (cdr prod)
                                   (lambda (x) (found-it (reverse x)))
                                   (lambda (x) (found-it x)))
                               (append
                                (let lp ((parent (hash-table/get seen prod #f)))
                                  (if (null? (cdr parent))
                                      '()
                                      (cons ((if (cdr prod)
                                                 (lambda (x) x)
                                                 inv-op)
                                             (car parent)) (lp (hash-table/get seen (cons (cdr parent) (cdr prod)) #f)))))
                                (let lp ((parent found-seen))
                                  (if (null? (cdr parent))
                                      '()
                                      (cons ((if (cdr prod)
                                                 inv-op
                                                 (lambda (x) x))
                                             (car parent)) (lp (hash-table/get seen (cons (cdr parent) (not (cdr prod))) #f))))))))
                            (lambda ()
                              (let ((the-ops (if (cdr prod) ops inv-ops)))
                                (for-each (lambda (op)
                                            (let ((prod2 (if (cdr prod)
                                                             ((group-operation g) op (car prod))
                                                             ((group-operation g) (car prod) op))))
                                              (if (eq? not-seen (hash-table/get seen (cons prod2 (cdr prod)) not-seen))
                                                  (begin
                                                    (hash-table/put! seen (cons prod2 (cdr prod))
                                                                     (cons op (car prod)))
                                                    (enqueue!/unsafe queue (cons prod2 (cdr prod)))))))
                                        the-ops))))
         (if (queue-empty? queue)
             #f
             (lp (dequeue!/unsafe queue))))))))

;;;
;;; 2x2x2 Rubik's cube
;;;

(define rubiks-parts
  '(
   flu luf ufl ; 0th cubie, front/left/up face
   fur urf rfu ; 1st cubie, front/up/right face
   fdl dlf lfd ; 2nd cubie, front/down/left face
   frd rdf dfr ; 3rd cubie, front/right/down face
   bul ulb lbu ; 4th cubie, back/up/left face
   bru rub ubr ; 5th cubie, back/right/up face
   bld ldb dbl ; 6th cubie, back/left/down face
   bdr drb rbd ; 7th cubie, back/down/right face
   ))

(define rubiks-parts-colors
  '(
   rgw gwr wrg
   rwb wbr brw
   ryg ygr gry
   rby byr yrb
   owg wgo gow
   obw bwo wob
   ogy gyo yog
   oyb ybo boy
   ))

(define (rubiks->permutation r)
  (map (lambda (e)
         (- (length r) (length (memq e rubiks-parts))))
       r))
(define (permutation->rubiks p)
  (map (lambda (e)
         (list-ref rubiks-parts e))
       p))
(define (rubiks->colors r)
  (cadr (assoc r (zip rubiks-parts rubiks-parts-colors))))
(define (colors->rubiks c)
  (cadr (assoc c (zip rubiks-parts-colors rubiks-parts))))

(define rubiks-identity
  (rubiks->permutation '(flu luf ufl fur urf rfu fdl dlf lfd frd rdf dfr bul ulb lbu bru rub ubr bld ldb dbl bdr drb rbd)))

(define rubiks-moves
  (list
   (rubiks->permutation '(fdl dlf lfd flu luf ufl frd rdf dfr fur urf rfu bul ulb lbu bru rub ubr bld ldb dbl bdr drb rbd))
   (perm-inverse (rubiks->permutation '(fdl dlf lfd flu luf ufl frd rdf dfr fur urf rfu bul ulb lbu bru rub ubr bld ldb dbl bdr drb rbd)))
   (rubiks->permutation '(ulb lbu bul fur urf rfu ufl flu luf frd rdf dfr dbl bld ldb bru rub ubr dlf lfd fdl bdr drb rbd))
   (perm-inverse (rubiks->permutation '(ulb lbu bul fur urf rfu ufl flu luf frd rdf dfr dbl bld ldb bru rub ubr dlf lfd fdl bdr drb rbd)))
   (rubiks->permutation '(rfu fur urf rub ubr bru fdl dlf lfd frd rdf dfr luf ufl flu lbu bul ulb bld ldb dbl bdr drb rbd))
   (perm-inverse (rubiks->permutation '(rfu fur urf rub ubr bru fdl dlf lfd frd rdf dfr luf ufl flu lbu bul ulb bld ldb dbl bdr drb rbd)))
   ))

(define rubiks-move-names
  '(F Fi L Li U Ui))

(define (rubiks-move->name move)
  (cadr (assoc move (zip rubiks-moves rubiks-move-names))))
(define (rubiks-name->move name)
  (cadr (assoc name (zip rubiks-move-names rubiks-moves))))

(define (rubiks-randomized)
  (let lp ((curr rubiks-identity)
           (to-go 30))
    (if (= 0 to-go)
        curr
        (lp (apply-perm (list-ref rubiks-moves (random (length rubiks-moves))) curr) (- to-go 1)))))

;;;
;;; Examples
;;;

;;
;; Chicken nugget problem
;;
; We want to get n chicken nuggets, but they only come in boxes of 5
; and 8.  What is the minimum number of boxes? (not really
; group-theory-specific)

#|
(pp (list-tabulate 100
                   (lambda (n)
                     (list n (solve-to-element (make-cyclic-group 100)
                                               n
                                               '(5 8)
                                               0)))))
|#


;;
;; Rubiks cube examples
;;

#|
(define r (make-generator-group rubiks-moves apply-perm))

(tag-inverse! r perm-inverse)

(solve-to-element r rubiks-identity rubiks-moves (rubiks-name->move 'F))

;Value 14: ((3 4 5 9 10 11 0 1 2 6 7 8 12 13 14 15 16 17 18 19 20 21 22 23))

(map rubiks-move->name (solve-to-element r rubiks-identity rubiks-moves (rubiks-name->move 'F)))

;Value 15: (fi)

(map rubiks-move->name (solve-to-element r rubiks-identity rubiks-moves (apply-perm (rubiks-name->move 'Ui) (rubiks-name->move 'F))))

;Value 16: (fi u)

(map rubiks-move->name (solve-to-element r rubiks-identity rubiks-moves (rubiks-randomized)))

;Value 11: (l f ui l fi u fi l fi u)

|#