;;;
;;; tc.scm --- todd-coxeter algorithm
;;;

(define (parse-relations relations sub-gens)
  (let ((mapping (make-eqv-hash-table))
        (count 0))
    (define (parse string)
      (let ((chars (string->list string)))
        (let loop ((chars chars) (result '()) (last #\0))
          (if (not (pair? chars))
              result
              (let ((char (car chars)))
                (cond ((equal? char #\^)
                       (loop (cdr chars) result last))
                      ((equal? char #\-)
                       (let ((negative (- 0(car result))))
                         (loop (cdr chars) (cons negative (cdr result)) negative)))
                      ((char->digit char)
                       (let digit-loop ((exponent (char->digit char)) (charss (cdr chars)))
                         (if (and (pair? charss) (char->digit (car charss)))
                             (digit-loop (+ (* 10 exponent) (char->digit (car charss))) (cdr charss))
                             (loop charss (append (make-list (- exponent 1) last) result) last))))
                      (else
                       (let ((code (hash-table/get mapping char #f)))
                         (if code
                             (loop (cdr chars) (cons code result) code)
                             (begin
                               (set! count (+ count 1))
                               (hash-table/put! mapping char count)
                               (loop (cdr chars) (cons count result) count))))
                             )))))))
    (define (parse-strings relations)
      (if (pair? relations)
          (cons (reverse (parse (car relations))) (parse-strings (cdr relations)))
          '()))
    (list count (parse-strings relations) (parse-strings sub-gens))))

; coset table - rows are cosets; c_ij = k => g_j*C_i = C_k under each element x,y,z,...,x^-1,...1 (the columns)
; relation table - artin; one for each relation, rows are cosets; begins and ends with coset number
; subgroup table - one table per generator of H; like coset table; columns are factors of that generator; starts and ends with 1
; H is coset 1

(define (tc relations sub-gens)
  (let* ((relations-data (parse-relations relations sub-gens))
         (num-generators  (car relations-data)) ; actually half the total, counting inverses
         (relations (cadr relations-data))
         (sub-gens (caddr relations-data)))

    (define (make-relation-tables relations)
      (if (pair? relations)
          (let ((table-row (make-vector (length (car relations)) #f)))
            (vector-set! table-row (- (vector-length table-row) 1) 0)
            (cons (vector table-row) (make-relation-tables (cdr relations))))
          '()))

    (define (table-ref table r c)
      (vector-ref (vector-ref table r) c))
    (define (table-set! table r c value)
      (vector-set! (vector-ref table r) c value))
    (define (delete-row! table r)
      (vector-append (vector-head table r) (vector-tail table (+ r 1))))
    
    (let ((coset-table (vector (make-vector (+ 1 (* 2 num-generators)) #f)))
          (relation-tables (make-relation-tables relations))
          (subgroup-tables (make-relation-tables sub-gens))
          (deductions-made #f))

      ; columns are indexed weirdly in this table
      (define (coset-table-ref c1 g)
        (if (not (and c1 g))
            #f
            (table-ref coset-table c1 (- num-generators g))))
      (define (coset-table-set! c1 g c2)
        (table-set! coset-table c1 (- num-generators g) c2))
      
      ; iterates over table, calling operation and changing the table value accordingly
      (define (iter table operation)
        (let loop1 ((i 0))
          (let ((row (vector-ref table i)))
            (let loop2 ((j 0))
              (let ((entry (vector-ref row j)))
                (operation i j entry))
                (if (< j (- (vector-length row) 1))
                    (loop2 (+ j 1)))))
          (if (< i (- (vector-length table) 1))
              (loop1 (+ i 1)))))

      ; fills in a slot of the coset-table, checks for coincidences
      (define (fill-in g c1 c2 k) ; g sends c1 to c2
        (let ((current (coset-table-ref c1 g)))
          (if (not current)
              (begin
                (set! deductions-made #t)
                (coset-table-set! c1 g c2)
                (coset-table-set! c2 (- 0 g) c1))
              (if (not (= current c2)) ; coincidence (two cosets are the same)
                  (let* ((correct (min current c2))
                         (wrong (max current c2))
                         (operation
                          (lambda (table)
                            (lambda (i j entry)
                              (cond ((not entry)) ; do nothing
                                    ((= entry wrong)
                                     (table-set! table i j correct))
                                    ((> entry wrong)
                                     (table-set! table i j (- entry 1))))))))
                    

                    (set! deductions-made #t) ; we need to loop again before we add a new coset
                    
                    ; delete bad rows and decrement higher-labelled cosets
                    (set! coset-table (delete-row! coset-table wrong))
                    (iter coset-table (operation coset-table))

                    (define (fix list delete?)
                      (let ((table (car list)))
                        (if delete?
                            (set! table (delete-row! table wrong)))
                        (iter table (operation table))
                        (set-car! list table))
                      (if (pair? (cdr list))
                          (fix (cdr list) delete?)))
                    
                    (fix relation-tables #t)
                    (fix subgroup-tables #f)
                    (k 'deleted-row) ; restart making deductions since table size changed
                    
                    ;(make-deductions)
                    )))))
      
      (define (make-deductions)
        (define (deduce table-list relations)
          (call-with-current-continuation
           (lambda (k)
             (let loop ((rel-num 0))
               (let ((table (list-ref table-list rel-num)))
                 (iter table
                       (lambda(i j entry)
                         (let* ((prev (if (= j 0) i (table-ref table i (- j 1))))
                                (next (if (= j (- (vector-length (vector-ref table i)) 1))
                                          #f
                                          (table-ref table i (+ j 1))))
                                (g-f (list-ref (list-ref relations rel-num) j))
                                (forward-action (coset-table-ref prev g-f))
                                (g-b (if (= j (- (vector-length (vector-ref table i)) 1))
                                         #f
                                         (- 0 (list-ref (list-ref relations rel-num) (+ j 1)))))
                                (backward-action (coset-table-ref next g-b)))
                           
                           (if (and prev entry)
                               (fill-in g-f prev entry k)
                               (if (and prev forward-action)
                                   (begin
                                     (set! deductions-made #t) 
                                     (table-set! table i j forward-action))))
                           (if (and entry next)
                               (fill-in g-b next entry k)
                               (if (and next backward-action)
                                   (begin
                                     (set! deductions-made #t)
                                     (table-set! table i j backward-action))))))))
               (if (< rel-num (- (length table-list) 1))
                   (loop (+ rel-num 1)))))))

        (deduce relation-tables relations)
        (deduce subgroup-tables sub-gens))

      ; returns first #f in coset table
      (define (first-unknown)
        (call-with-current-continuation
         (lambda (k)
           (let loop1 ((i 0))
             (let loop2 ((j 1))
               (if (not (coset-table-ref i j))
                   (k (list i j)))
               (if (not (coset-table-ref i (- 0 j)))
                   (k (list i (- 0 j))))
               (if (< j num-generators)
                   (loop2 (+ j 1))))
             (if (< i (- (vector-length coset-table) 1))
                 (loop1 (+ i 1))))
           (k #f))))
                    
                
      (define (add-coset c1 g)
        (let ((c (vector-length coset-table))) ; name of new coset (cosets are 0 indexed)
          ; adds row to coset-table
          (set! coset-table (vector-grow coset-table (+ c 1)))
          (vector-set! coset-table c (make-vector (+ 1 (* 2 num-generators)) #f))
          (coset-table-set! c1 g c)
          (coset-table-set! c (- 0 g) c1)
          ; adds row to each relation  table
          (define (add-row list)
            (let ((table (vector-grow (car list) (+ c 1))))
              (vector-set! table c (make-vector (vector-length (vector-ref table 0)) #f))
              (vector-set! (vector-ref table c) (- (vector-length (vector-ref table 0)) 1) c)
              (set-car! list table))
            (if (pair? (cdr list))
                (add-row (cdr list))))
          (add-row relation-tables)))
              


      (define (make-group-generators)
        (let loop1 ((j 1))
          (if (> j num-generators)
              '()
              (cons
               (let loop2 ((i 0))
                 (if (= i (vector-length coset-table))
                     '()
                     (cons
                      (coset-table-ref i j)
                      (loop2 (+ i 1)))))
               (loop1 (+ j 1))))))               
          

      (let main-loop ()
        (let deduction-loop ()
          (set! deductions-made #f)
          (make-deductions)
          (if deductions-made
              (deduction-loop)))
        (let ((empty-slot (first-unknown)))
          (if empty-slot
            (begin
              (add-coset (car empty-slot) (cadr empty-slot))
              (main-loop))
            (make-group-generators)))))))