;;;
;;; threaded-engine.scm --- a multi-threaded execution engine for
;;; uninspired solution-making
;;;

(load "conspire.scm")

;;;
;;; Solution pipes
;;;

;; A solution pipe is a device which waits for a solver to write to
;; it, thus allowing multiple threads to wait for the same answer.

;; It assumes each thread will ultimately yield the same answer (if
;; not unknown).


;; A slight change so we can store where the answer is supposed to go
;; (so we can kill a thread correctly).
(define-record-type conspire:thread
    (conspire:make-new-thread continuation)
    conspire:thread?
  (continuation conspire:continuation
		conspire:set-continuation!)
  (solution-pipe conspire:solution-pipe
               conspire:set-solution-pipe!))

;; Returned from threads which don't actually know the answer in any case
(define unknown (list 'unknown))
(define (is-unknown? x) (eq? unknown x))

;; Kept in the pipe until a real answer is found.
(define no-solution-yet (list 'no-solution-yet))
(define (has-solution? x) (not (eq? x no-solution-yet)))

;; Makes a "pipe" which waits until a non-unknown value comes in, or
;; until each thread comes in with unknown
(define (make-solution-pipe num-threads)
  (let* ((answer no-solution-yet)
         (lock (conspire:make-lock))
         (owner-threads (if conspire:running?
                            (list *running-thread*)
                            '())))
    (define pipe-in
      (lambda (value)
        (conspire:acquire-lock lock)
        (set! num-threads (- num-threads 1))
        (if (not (is-unknown? value))
            (begin
              (if (and (has-solution? answer) (not (equal? value answer)) (not (and (procedure? value) (procedure? answer))))
                  (error "Inconsistent solutions in pipe" value answer))
              (set! answer value)))
        (if (and (not (has-solution? answer)) (= 0 num-threads))
            (if (not (null? owner-threads))
                (for-each
                 (lambda (thread)
                   ((pipe-writer (conspire:solution-pipe thread)) unknown)
                   (conspire:kill-threads (list thread)))
                 owner-threads)
                (error "pipe: no more threads to give solutions")))
        (conspire:unlock lock)))
    (define pipe-out
      (lambda ()
        (if (has-solution? answer)
            answer
            (begin
              (conspire:switch-threads
               (lambda () (has-solution? answer)))
              (pipe-out)))))
    (define add-listener
      (lambda (thread)
        (conspire:acquire-lock lock)
        (set! owner-threads (cons thread owner-threads))
        (conspire:unlock lock)))
    (list pipe-in pipe-out add-listener)))
(define (pipe-reader pipe)
  (cadr pipe))
(define (pipe-writer pipe)
  (car pipe))
(define (pipe-add-listener! pipe listener)
  ((caddr pipe) listener))

;;;
;;; Threaded engines
;;;

;; This is so we can keep everything within one time sharing system
(define *run-maybe-with-time-sharing-conspiracy-before?* #f)
(define (maybe-with-time-sharing-conspiracy thunk)
  (call-with-current-continuation ; a hack to get around register-timer-event
   (lambda (continuation)
     (if (not *run-maybe-with-time-sharing-conspiracy-before?*)
         (create-thread continuation
                        (lambda ()
                          (let lp ()
                            (set! *run-maybe-with-time-sharing-conspiracy-before?* #t)
                            (sleep-current-thread 10)
                            (lp)))))
     (if conspire:running?  ; and the regularly-scheduled programming
         (thunk)
         (with-time-sharing-conspiracy thunk)))))

(define *threaded-solvers* (make-eq-hash-table))
(define *threaded-solvers-arity* (make-eq-hash-table))

;; The solutions from the threaded solvers are memoized, and different
;; threads which are requesting the same solver at the same time will
;; wait upon the same solver.
(define *current-jobs* (make-eq-hash-table))
(define (add-job-pipe! engine args pipe-reader)
  (let ((job-hash-table (hash-table/get *current-jobs* engine #f)))
    (if job-hash-table
        (hash-table/put! job-hash-table args pipe-reader)
        (begin
          (let ((table ((weak-hash-table/constructor equal-hash-mod equal? #t))))
            (hash-table/put! table args pipe-reader)
            (hash-table/put! *current-jobs* engine table))))))

;; Makes a threaded solver which runs, in parallel, each of the
;; solvers added to it.  Each solver must be of the given arity.  Once
;; completed, the procedure when-done, which takes in as arguments the
;; answer and the arguments where applied to the solvers, is run if it
;; isn't #f (to perhaps to store the found information into a
;; convenient data structure for later retrieval).
(define (make-threaded-solver when-done arity)
  (letrec ((engine (lambda args
                     (let ((solver-pipe (get-solver-pipe engine args)))
                       (if solver-pipe
                           (begin
                             (if conspire:running?
                                 (pipe-add-listener! solver-pipe *running-thread*))
                             ((pipe-reader solver-pipe)))
                           (let* ((solvers (hash-table/get *threaded-solvers* engine '()))
                                  (pipe (make-solution-pipe (length solvers))))
                             (add-job-pipe! engine args pipe)
                             (let ((ans (execute-threaded-solver solvers args pipe)))
                               (if when-done
                                   (apply when-done ans args))
                               ans)))))))
    (hash-table/put! *threaded-solvers* engine '())
    (hash-table/put! *threaded-solvers-arity* engine arity)
    engine))

;; Helper function to attach a bunch of thunks to a pipe and wait on
;; the pipe for the first result.
(define (execute-threaded-solver solvers args pipe)
  (maybe-with-time-sharing-conspiracy
   (lambda ()
     (let* ((threads
             (map (lambda (solver)
                    (let ((thread
                           (conspire:make-thread
                            conspire:runnable
                            (lambda ()
                              (let ((ret (apply solver args)))
                                ((pipe-writer pipe) ret))))))
                      (conspire:set-solution-pipe! thread pipe)
                      thread)) ;; do it.
                  solvers))
            (ret ((pipe-reader pipe))))
       (conspire:kill-threads threads)
       ret))))

;; Runs a bunch of thunks and waits for the first result.  Handles
;; unknown (failed) results nicely.
(define (run-threads . threads)
  (execute-threaded-solver threads '() (make-solution-pipe (length threads))))

(define (get-solver-pipe engine args)
  (let ((engine-pipes (hash-table/get *current-jobs* engine #f)))
    (and engine-pipes (hash-table/get engine-pipes args #f))))

;; Adds a procedure called solver to the given engine.  It must return
;; the correct value or the value unknown.  It will be memoized, so it
;; must be functional.
(define (add-threaded-solver engine solver)
  (if (not (hash-table/get *threaded-solvers-arity* engine #f))
      (error "add-threaded-solver needs an existant engine to add a solver to"))
  (if (not (procedure-arity-valid? solver
                                   (hash-table/get *threaded-solvers-arity* engine 0)))
      (error "solver must be same arity as engine"))
  (hash-table/put! *threaded-solvers*
                 engine (cons solver
                              (hash-table/get *threaded-solvers* engine '())))
  engine)

(define defsolver add-threaded-solver)

;;;
;;; a test
;;;
#|
(define test-solver (make-threaded-solver #f 1))

(add-threaded-solver test-solver (lambda (n)
                                   (let lp ((i 0))
                                     (if (= i n)
                                         (begin (display "done a") i)
                                         (lp (+ i 1))))))
(add-threaded-solver test-solver (lambda (n)
                                   (let lp ((i 1000000))
                                     (if (= i n)
                                         (begin (display "done b") i)
                                         (lp (- i 1))))))

(define test-solver2 (make-threaded-solver #f 0))
(add-threaded-solver test-solver2 (lambda ()
                                    (display "<?")
                                    (test-solver 100)
                                    (display "?>")
                                    5))
(add-threaded-solver test-solver2 (lambda ()
                                    (display "{-")
                                    (test-solver 100)
                                    (display "-}")
                                    5))
|#

;; Example:
;; 1 ]=> (test-solver 10)
;; done a
;; ;Value: 10

;; 1 ]=> (test-solver 999999)
;; done b
;; ;Value: 999999

#|
(define unknown-test (make-threaded-solver #f 0))
(add-threaded-solver unknown-test (lambda ()
                                    unknown))
;; Example:
;; 1 ]=> (unknown-test)
;; ;pipe: no more threads to give solutions
;; ;To continue, call RESTART with an option number:
;; ; (RESTART 1) => Return to read-eval-print level 1.

(define unknown-test2 (make-threaded-solver #f 0))
(add-threaded-solver unknown-test2 (lambda ()
                                     (let lp ((i 1000000))
                                       (if (= i 0)
                                           13
                                           (lp (- i 1))))))
(add-threaded-solver unknown-test2 (lambda ()
                                     (let ((a (unknown-test)))
                                       (+ a 1))))

;; Example:
;; 1 ]=> (unknown-test2)
;; ;Value: 13

|#